"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const model = require("./model");
exports.MINIMAL = 'MINIMAL';
exports.MAXIMAL = 'MAXIMAL';
class DataLevelModel extends model.Model {
    toJSON(options) {
        let obj = this.toObject();
        if (options && options.level) {
            const fields = (this.schema.dataLevel && this.schema.dataLevel.levelMap[options.level]
                || []);
            obj = _.pick(obj, fields);
        }
        return obj;
    }
}
exports.DataLevelModel = DataLevelModel;
DataLevelModel.Plugin({
    fn: dataLevelPlugin,
    priority: 100,
});
function dataLevelPlugin(schema, options) {
    if (schema.dataLevel == null) {
        schema.dataLevel = {
            levelMap: {},
        };
    }
    // schema.levels = [ MINIMAL ].concat(options.levels || []);
    addToLevelMap(schema, levelPaths(schema));
    for (let name of model.preQueries) {
        schema.pre(name, modifyProjection);
    }
}
function modifyProjection(next) {
    let schema = this.schema;
    let level = (this.options.level ||
        schema.options.dataLevel && schema.options.dataLevel.default);
    if (level) {
        let fields = schema.dataLevel.levelMap[level] || [];
        if (this._fields == null) {
            this._fields = {};
        }
        for (let field of fields) {
            this._fields[field] = 1;
        }
    }
    next();
}
function addToLevelMap(schema, levelPaths) {
    if (schema.parentSchema) {
        addToLevelMap(schema.parentSchema, levelPaths);
    }
    let dataLevelOptionsLevels = (schema.options.dataLevel && schema.options.dataLevel.levels || []);
    let levelMap = schema.dataLevel.levelMap;
    for (let { path, level } of levelPaths) {
        for (let levelIndex = (level === exports.MINIMAL ? 0 : dataLevelOptionsLevels.indexOf(level)); levelIndex >= 0 && levelIndex < dataLevelOptionsLevels.length; levelIndex++) {
            let cLevel = dataLevelOptionsLevels[levelIndex];
            levelMap[cLevel] = _.union(levelMap[cLevel], [path]);
        }
        if (level === exports.MINIMAL) {
            levelMap[exports.MINIMAL] = _.union(levelMap[exports.MINIMAL], [path]);
        }
        levelMap[exports.MAXIMAL] = _.union(levelMap[exports.MAXIMAL], [path]);
    }
}
function levelPaths(schema) {
    let res = [];
    schema.eachPath((pathname, schemaType) => {
        res.push({
            path: pathname,
            level: schemaType.options.level || exports.MINIMAL,
        });
    });
    return res;
}

//# sourceMappingURL=data-level.js.map
