"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
const _ = require("underscore");
const mongoose_1 = require("mongoose");
/**
 * Wrapped Model from mongoose.Model.
 */
class Model {
    static Config(config) {
        if (this.hasOwnProperty('$CONFIG')) {
            _.extend(this.$CONFIG, config);
        }
        else {
            this.$CONFIG = config;
        }
        return this;
    }
    static Schema(schema) {
        if (this.hasOwnProperty('$SCHEMA')) {
            _.extend(this.$SCHEMA, schema);
        }
        else {
            this.$SCHEMA = schema;
        }
        return this;
    }
    static Pre(pre) {
        if (this.hasOwnProperty('$PRES')) {
            this.$PRES.push(pre);
        }
        else {
            this.$PRES = [pre];
        }
        return this;
    }
    static Post(post) {
        if (this.hasOwnProperty('$POSTS')) {
            this.$POSTS.push(post);
        }
        else {
            this.$POSTS = [post];
        }
        return this;
    }
    static Virtual(virtual) {
        if (this.hasOwnProperty('$VIRTUALS')) {
            this.$VIRTUALS.push(virtual);
        }
        else {
            this.$VIRTUALS = [virtual];
        }
        return this;
    }
    static Plugin(plugin) {
        if (this.hasOwnProperty('$PLUGINS')) {
            this.$PLUGINS.push(plugin);
        }
        else {
            this.$PLUGINS = [plugin];
        }
        return this;
    }
    static Index(index) {
        if (this.hasOwnProperty('$INDEXES')) {
            this.$INDEXES.push(index);
        }
        else {
            this.$INDEXES = [index];
        }
        return this;
    }
    static Mixin(model) {
        if (this.hasOwnProperty('$MIXINS')) {
            this.$MIXINS.push(model);
        }
        else {
            this.$MIXINS = [model];
        }
        return this;
    }
    static Validate(validate) {
        if (this.hasOwnProperty('$VALIDATES')) {
            this.$VALIDATES.push(validate);
        }
        else {
            this.$VALIDATES = [validate];
        }
        return this;
    }
    static _$modifySchema(discriminatorKey) {
        if (!this.hasOwnProperty('$SCHEMA')) {
            this.$SCHEMA = {};
        }
        const uniqueFields = [];
        for (const name in this.$SCHEMA) {
            const opt = this.$SCHEMA[name];
            if (opt.unique) {
                opt.unique = false;
                uniqueFields.push(name);
            }
        }
        for (const field of uniqueFields) {
            const fieldUniqueKey = `${field}_unique`;
            const fields = {};
            fields[discriminatorKey] = 1;
            fields[field] = 1;
            fields[fieldUniqueKey] = '2dsphere';
            this.$SCHEMA[fieldUniqueKey] = {
                type: PointSchema,
                // api:        AUTOGEN,
                default: PointSchema,
            };
            this.Index({
                fields,
                options: {
                    unique: true,
                },
            });
        }
        this.$SCHEMA[discriminatorKey] = { type: String };
        this.Pre({
            name: 'save',
            fn: setDiscriminatorKey,
        });
        for (const name of exports.preQueries) {
            this.Pre({ name, fn: patchDiscriminatorKey });
        }
        function setDiscriminatorKey(next) {
            if (this[discriminatorKey] == null) {
                this[discriminatorKey] = this.constructor.modelName;
            }
            next();
        }
        function patchDiscriminatorKey() {
            this._conditions[discriminatorKey] = this.model.modelName;
        }
    }
    static $mongooseOptions() {
        this._$initialize();
        if (this._mongooseOptions.initialized || this === Model) {
            return this._mongooseOptions;
        }
        const superClass = this.__proto__;
        const superOptions = (superClass.$mongooseOptions ? superClass.$mongooseOptions() : {
            initialized: false,
        });
        if (superOptions.config.discriminatorKey) {
            this._$modifySchema(superOptions.config.discriminatorKey);
        }
        const mixinModels = this.hasOwnProperty('$MIXINS') ? this.$MIXINS : [];
        const mixinOptions = _.map(mixinModels, (model) => model.$mongooseOptions());
        const mixinSchemas = _.map(mixinOptions, (opt) => opt.schema);
        const decoSchema = Reflect.getOwnMetadata(schemaKey, this.prototype) || {};
        const flattenSchemas = _.flatten([{}, mixinSchemas, superOptions.schema, this.$SCHEMA, decoSchema]);
        this._mongooseOptions.schema = _.extend.apply(null, flattenSchemas);
        const mixinConfigs = _.map(mixinOptions, (opt) => opt.config);
        const decoConfigs = Reflect.getOwnMetadata(configKey, this.prototype) || {};
        const flattenConfigs = _.flatten([{}, mixinConfigs, superOptions.config, this.$CONFIG, decoConfigs]);
        this._mongooseOptions.config = _.extend.apply(null, flattenConfigs);
        const mixinPres = _.map(mixinOptions, (opt) => opt.pres);
        const decoPres = Reflect.getOwnMetadata(preKey, this.prototype) || [];
        this._mongooseOptions.pres = _.union(_.flatten([
            mixinPres, superOptions.pres || [],
            this.hasOwnProperty('$PRES') ? this.$PRES : [],
            decoPres,
        ]));
        const mixinPosts = _.map(mixinOptions, (opt) => opt.posts);
        const decoPosts = Reflect.getOwnMetadata(postKey, this.prototype) || [];
        this._mongooseOptions.posts = _.union(_.flatten([
            mixinPosts, superOptions.posts || [],
            this.hasOwnProperty('$POSTS') ? this.$POSTS : [],
            decoPosts,
        ]));
        const mixinVirtuals = _.map(mixinOptions, (opt) => opt.virtuals);
        this._mongooseOptions.virtuals = _.union(_.flatten([
            mixinVirtuals, superOptions.virtuals || [],
            this.hasOwnProperty('$VIRTUALS') ? this.$VIRTUALS : [],
        ]));
        const mixinValidates = _.map(mixinOptions, (opt) => opt.validates);
        const decoValidates = Reflect.getOwnMetadata(validateKey, this.prototype);
        this._mongooseOptions.validates = _.union(_.flatten([
            mixinValidates, superOptions.validates || [],
            this.hasOwnProperty('$VALIDATES') ? this.$VALIDATES : [],
            decoValidates || [],
        ]));
        const mixinPlugins = _.map(mixinOptions, (opt) => opt.plugins);
        const decoPlugins = Reflect.getOwnMetadata(pluginKey, this.prototype) || [];
        this._mongooseOptions.plugins = _.union(_.flatten([
            mixinPlugins, superOptions.plugins || [],
            this.hasOwnProperty('$PLUGINS') ? this.$PLUGINS : [],
            decoPlugins,
        ]));
        this._mongooseOptions.plugins = _.sortBy(this._mongooseOptions.plugins, (plugin) => plugin.priority);
        const mixinIndexes = _.map(mixinOptions, (opt) => opt.indexes);
        const decoIndexes = Reflect.getOwnMetadata(indexKey, this.prototype) || [];
        this._mongooseOptions.indexes = _.union(_.flatten([
            mixinIndexes, superOptions.indexes || [],
            this.hasOwnProperty('$INDEXES') ? this.$INDEXES : [],
            decoIndexes,
        ]));
        const mixinMethods = _.map(mixinOptions, (opt) => opt.methods);
        this._mongooseOptions.methods = _.union(_.flatten([
            mixinMethods, superOptions.methods || [],
        ]));
        const mixinStatics = _.map(mixinOptions, (opt) => opt.statics);
        this._mongooseOptions.statics = _.union(_.flatten([
            mixinStatics, superOptions.statics || [],
        ]));
        for (const name of Object.getOwnPropertyNames(this.prototype)) {
            if (name === 'constructor') {
                continue;
            }
            const descriptor = Object.getOwnPropertyDescriptor(this.prototype, name);
            if (descriptor.value && _.isFunction(descriptor.value)) {
                this._mongooseOptions.methods.push({
                    name,
                    fn: descriptor.value,
                });
            }
            if (descriptor.get || descriptor.set) {
                this._mongooseOptions.virtuals.push({
                    name,
                    get: descriptor.get,
                    set: descriptor.set,
                });
            }
        }
        for (const name of Object.getOwnPropertyNames(this)) {
            const descriptor = Object.getOwnPropertyDescriptor(this, name);
            if (STATIC_FILTER_NAMES.indexOf(name) >= 0) {
                continue;
            }
            if (descriptor.value) {
                this._mongooseOptions.statics.push({
                    name,
                    fn: descriptor.value,
                });
            }
        }
        const mongooseSchema = new mongoose_1.Schema(this._mongooseOptions.schema, _.clone(this._mongooseOptions.config));
        mongooseSchema.parentSchema = superOptions.mongooseSchema;
        for (const pre of this._mongooseOptions.pres) {
            pre.parallel = pre.parallel || false;
            mongooseSchema.pre(pre.name, pre.parallel, pre.fn, pre.errorCb);
        }
        for (const post of this._mongooseOptions.posts) {
            mongooseSchema.post(post.name, post.fn);
        }
        for (const virtual of this._mongooseOptions.virtuals) {
            let v = mongooseSchema.virtual(virtual.name);
            if (virtual.get) {
                v = v.get(virtual.get);
            }
            if (virtual.set) {
                v = v.set(virtual.set);
            }
        }
        for (const method of this._mongooseOptions.methods) {
            mongooseSchema.methods[method.name] = method.fn;
        }
        for (const method of this._mongooseOptions.statics) {
            mongooseSchema.statics[method.name] = method.fn;
        }
        for (const plugin of this._mongooseOptions.plugins) {
            const options = _.extend({}, this._mongooseOptions.config, plugin.options);
            mongooseSchema.plugin(plugin.fn, options);
        }
        for (const index of this._mongooseOptions.indexes) {
            mongooseSchema.index(index);
        }
        for (const validate of this._mongooseOptions.validates) {
            mongooseSchema
                .path(validate.path)
                .validate(validate.fn, validate.errorMsg, validate.type);
        }
        this._mongooseOptions.mongooseSchema = mongooseSchema;
        this._mongooseOptions.initialized = true;
        return this._mongooseOptions;
    }
    static $register(mongooseInstance) {
        if (!mongooseInstance) {
            mongooseInstance = require('mongoose');
        }
        const model = mongooseInstance.model(this.name, this.$mongooseOptions().mongooseSchema);
        return model;
    }
    static _$initialize() {
        if (this.hasOwnProperty('_mongooseOptions')) {
            return;
        }
        this._mongooseOptions = {
            initialized: false,
            config: {},
        };
    }
    static cast() {
        return this;
    }
}
exports.Model = Model;
const schemaKey = Symbol('sbase:schema');
const configKey = Symbol('sbase:config');
const pluginKey = Symbol('sbase:plugin');
const preKey = Symbol('sbase:pre');
const postKey = Symbol('sbase:post');
const indexKey = Symbol('sbase:index');
const validateKey = Symbol('sbase:validate');
function Field(schema) {
    return (target, propertyName) => {
        const schemas = Reflect.getOwnMetadata(schemaKey, target) || {};
        schemas[propertyName] = schema;
        Reflect.defineMetadata(schemaKey, schemas, target);
    };
}
exports.Field = Field;
function Config(config) {
    return (constructor) => {
        const configs = Reflect.getOwnMetadata(configKey, constructor.prototype) || {};
        _.extend(configs, config);
        Reflect.defineMetadata(configKey, configs, constructor.prototype);
    };
}
exports.Config = Config;
function Plugin(plugin) {
    return (constructor) => {
        const plugins = Reflect.getOwnMetadata(pluginKey, constructor.prototype) || [];
        plugins.push(plugin);
        Reflect.defineMetadata(pluginKey, plugins, constructor.prototype);
    };
}
exports.Plugin = Plugin;
function Pre(pre) {
    return (constructor) => {
        const pres = Reflect.getOwnMetadata(preKey, constructor.prototype) || [];
        pres.push(pre);
        Reflect.defineMetadata(preKey, pres, constructor.prototype);
    };
}
exports.Pre = Pre;
function Pres(names, pre) {
    return (constructor) => {
        const pres = Reflect.getOwnMetadata(preKey, constructor.prototype) || [];
        for (const name of names) {
            pres.push(_.extend({ name }, pre));
        }
        Reflect.defineMetadata(preKey, pres, constructor.prototype);
    };
}
exports.Pres = Pres;
function Post(post) {
    return (constructor) => {
        const posts = Reflect.getOwnMetadata(postKey, constructor.prototype) || [];
        posts.push(post);
        Reflect.defineMetadata(postKey, posts, constructor.prototype);
    };
}
exports.Post = Post;
function Posts(names, post) {
    return (constructor) => {
        const posts = Reflect.getOwnMetadata(postKey, constructor.prototype) || [];
        for (const name of names) {
            posts.push(_.extend({ name }, post));
        }
        Reflect.defineMetadata(postKey, posts, constructor.prototype);
    };
}
exports.Posts = Posts;
function Index(index) {
    return (constructor) => {
        const indexes = Reflect.getOwnMetadata(indexKey, constructor.prototype) || [];
        indexes.push(index);
        Reflect.defineMetadata(indexKey, indexes, constructor.prototype);
    };
}
exports.Index = Index;
function Validate(validate) {
    return (constructor) => {
        const validates = Reflect.getOwnMetadata(validateKey, constructor.prototype) || [];
        validates.push(validate);
        Reflect.defineMetadata(validateKey, validates, constructor.prototype);
    };
}
exports.Validate = Validate;
class NativeError extends global.Error {
}
exports.NativeError = NativeError;
exports.preQueries = [
    'find', 'findOne', 'count', 'findOneAndUpdate', 'findOneAndRemove', 'update',
];
const PointSchema = new mongoose_1.Schema({
    type: {
        type: String,
        enum: ['Point'],
        default: 'Point',
    },
    coordinates: {
        type: [Number],
        default: [0, 0],
    },
});
const STATIC_FILTER_NAMES = [
    'name', 'prototype', '_mongooseOptions', '$PLUGINS', '$PRES', '$SCHEMA',
    '$MIXINS', '$CONFIG', '$INDEXES',
];
// Patch MModel.prototype.init
const _init = mongoose_1.Model.prototype.init;
mongoose_1.Model.cast = Model.cast;
mongoose_1.Model.prototype.init = function (doc, query, fn) {
    const discriminatorKey = this.schema.options.discriminatorKey;
    const type = doc[discriminatorKey];
    if (type) {
        const model = this.db.model(type);
        if (model) {
            this.schema = model.schema;
            this.__proto__ = model.prototype;
            _init.call(this, doc, query);
            if (fn) {
                fn(null);
            }
            return this;
        }
    }
    _init.call(this, doc, query, fn);
};

//# sourceMappingURL=model.js.map
