"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const LRU = require("lru-cache");
const logger = require("@nodeswork/logger");
require("@nodeswork/utils/dist/promise");
const LOG = logger.getLogger();
class SocketRpcClient {
    constructor(socket) {
        this.socket = socket;
        this.$requsetId = 0;
    }
    async call(name, args, timeoutMillis) {
        const request = {
            requestId: ++this.$requsetId,
            responseEventName: this.responseEventName,
            timeoutMillis,
            name,
            args,
        };
        LOG.debug('Send rpc request', request);
        this.socket.emit(this.requestEventName, request);
        const promise = new Promise((resolve, reject) => {
            this.$cache.set(this.$requsetId, [resolve, reject], timeoutMillis);
        }).timeout(timeoutMillis);
        return promise;
    }
}
exports.SocketRpcClient = SocketRpcClient;
function remote(options = {}) {
    return (target, propertyKey, descriptor) => {
        descriptor.value = function () {
            return this.call(propertyKey, Array.prototype.slice.call(arguments), options.timeoutMillis || this.timeoutMillis);
        };
    };
}
exports.remote = remote;
function socketRpcClient(options = {}) {
    return (constructor) => {
        if (!(constructor.prototype instanceof SocketRpcClient)) {
            throw new Error(`${constructor.name} is not a subclass of SocketRpcClient`);
        }
        return class extends constructor {
            constructor() {
                super(...arguments);
                this.timeoutMillis = options.timeoutMillis || 1000;
                this.requestEventName = `${options.eventNamePrefix || 'socket-rpc'}.request`;
                this.responseEventName = `${options.eventNamePrefix || 'socket-rpc'}.response`;
                this.$cache = new LRU();
                this.___ = this.socket.on(this.responseEventName, (resp) => {
                    LOG.debug('Receive rpc response', resp);
                    const cached = this.$cache.get(resp.requestId);
                    if (cached == null) {
                        LOG.error('Request is already timedout', resp);
                        return;
                    }
                    const [resolve, reject] = cached;
                    if (resp.error != null) {
                        reject(resp.error);
                    }
                    else {
                        resolve(resp.result);
                    }
                });
            }
        };
    };
}
exports.socketRpcClient = socketRpcClient;

//# sourceMappingURL=client.js.map
