"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@nodeswork/utils");
const logger = require("@nodeswork/logger");
const LOG = logger.getLogger();
const UNKOWN_METHOD_ERROR = new utils_1.NodesworkError('unkown method').toJSON();
function socketRpcHost(socket, target, eventNamePrefix) {
    socket.on(`${eventNamePrefix}.request`, async (request) => {
        LOG.debug('Receive request', request);
        const fn = target[request.name];
        let result;
        let error;
        if (fn == null) {
            error = UNKOWN_METHOD_ERROR;
        }
        else {
            try {
                result = await fn.apply(target, request.args);
            }
            catch (e) {
                error = utils_1.NodesworkError.cast(e).toJSON({
                    cause: true,
                    stack: true,
                });
            }
        }
        const response = {
            requestId: request.requestId,
            result,
            error,
        };
        LOG.debug('Send response', response);
        socket.emit(request.responseEventName, response);
    });
}
exports.socketRpcHost = socketRpcHost;

//# sourceMappingURL=host.js.map
